from bson.objectid import ObjectId
from functools import partial
from uengine import ctx
from uengine.errors import ApiError, NotFound
from uengine.utils import resolve_id

from .storable_model import StorableModel


class MissingShardId(ApiError):
    status_code = 500


# pylint: disable=arguments-differ
class ShardedModel(StorableModel):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._shard_id = None
        if "shard_id" in kwargs:
            self._shard_id = kwargs["shard_id"]
        if not self.is_new and self._shard_id is None:
            from traceback import print_stack
            print_stack()
            raise MissingShardId(
                "ShardedModel from database with missing shard_id - this must be a bug")

    @property
    def _db(self):
        return ctx.db.shards[self._shard_id]

    def save(self, skip_callback=False):
        if self._shard_id is None:
            raise MissingShardId(
                "ShardedModel must have shard_id set before save")
        super().save(skip_callback)

    def _refetch_from_db(self):
        return self.find_one(self._shard_id, {"_id": self._id})

    @classmethod
    def _get_possible_databases(cls):
        return list(ctx.db.shards.values())

    @classmethod
    def find(cls, shard_id, query=None, **kwargs):
        if not query:
            query = {}
        return ctx.db.get_shard(shard_id).get_objs(
            cls.from_data,
            cls.collection,
            cls._preprocess_query(query),
            **kwargs
        )

    @classmethod
    def aggregate(cls, shard_id, pipeline, query=None, **kwargs):
        if not query:
            query = {}
        pipeline = [{"$match": cls._preprocess_query(query)}] + pipeline
        return ctx.db.get_shard(shard_id).get_aggregated(cls.collection, pipeline, **kwargs)

    @classmethod
    def find_projected(cls, shard_id, query=None, projection=('_id',), **kwargs):
        if not query:
            query = {}
        return ctx.db.get_shard(shard_id).get_objs_projected(
            cls.collection,
            cls._preprocess_query(query),
            projection=projection,
            **kwargs
        )

    @classmethod
    def find_one(cls, shard_id, query, **kwargs):
        return ctx.db.get_shard(shard_id).get_obj(
            cls.from_data,
            cls.collection,
            cls._preprocess_query(query),
            **kwargs
        )

    @classmethod
    def get(cls, shard_id, expression, raise_if_none=None):
        if expression is None:
            return None
        expression = resolve_id(expression)
        if isinstance(expression, ObjectId):
            query = {"_id": expression}
        else:
            expression = str(expression)
            query = {cls.KEY_FIELD: expression}
        res = cls.find_one(shard_id, query)
        if res is None and raise_if_none is not None:
            if isinstance(raise_if_none, Exception):
                raise raise_if_none
            else:
                raise NotFound(f"{cls.__name__} not found")
        return res

    @classmethod
    def cache_get(cls, shard_id, expression, raise_if_none=None):
        if expression is None:
            return None
        cache_key = f"{cls.collection}.{shard_id}.{expression}"
        getter = partial(cls.get, shard_id, expression, raise_if_none)
        constructor = partial(cls.from_data, shard_id=shard_id)
        obj = cls._cache_get(cache_key, getter, constructor)
        obj.shard_id = shard_id
        return obj

    def invalidate(self):
        cache_key_id = f"{self.collection}.{self._shard_id}.{self._id}"
        cache_key_keyfield = None
        if self.KEY_FIELD is not None and self.KEY_FIELD != "_id":
            cache_key_keyfield = f"{self.collection}.{self._shard_id}.{getattr(self, self.KEY_FIELD)}"

        return self._invalidate(cache_key_id, cache_key_keyfield)

    @classmethod
    def destroy_all(cls, shard_id):
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        ctx.db.get_shard(shard_id).delete_query(
            cls.collection, cls._preprocess_query({}))

    @classmethod
    def destroy_many(cls, shard_id, query):
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        ctx.db.get_shard(shard_id).delete_query(
            cls.collection, cls._preprocess_query(query))

    @classmethod
    def update_many(cls, shard_id, query, attrs):
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        ctx.db.get_shard(shard_id).update_query(
            cls.collection, cls._preprocess_query(query), attrs)



# pylint: disable=protected-access

from uengine import ctx
from uengine.models.sharded_model import ShardedModel, MissingShardId
from .mongo_mock import MongoMockTest

CALLABLE_DEFAULT_VALUE = 4


def callable_default():
    return CALLABLE_DEFAULT_VALUE


class TestModel(ShardedModel):
    FIELDS = (
        '_id',
        'field1',
        'field2',
        'field3',
        'callable_default_field'
    )

    DEFAULTS = {
        'field1': 'default_value',
        'field3': 'required_default_value',
        'callable_default_field': callable_default
    }

    REQUIRED_FIELDS = (
        'field2',
        'field3',
    )

    REJECTED_FIELDS = (
        'field1',
    )

    INDEXES = (
        "field1",
    )


class TestShardedModel(MongoMockTest):

    def setUp(self):
        super().setUp()
        for shard_id in ctx.db.shards:
            TestModel.destroy_all(shard_id)

    def tearDown(self):
        for shard_id in ctx.db.shards:
            TestModel.destroy_all(shard_id)
        super().tearDown()

    def test_init(self):
        model = TestModel(field1="value")
        self.assertEqual(model.field1, "value")
        model._before_delete()
        model._before_save()

    def test_shard(self):
        model = TestModel(field2="value")
        self.assertIsNone(model._shard_id)
        self.assertRaises(MissingShardId, model.save)

        shard_id = ctx.db.rw_shards[0]
        model = TestModel(shard_id=shard_id, field2="value")
        self.assertEqual(model._shard_id, shard_id)
        model.save()



import os

from lazy_object_proxy import Proxy
from uengine.base import Base
from uengine import ctx

from .controllers.main import gen_main_ctrl


def get_version():
    cur_dir = os.path.dirname(__file__)
    version_file = os.path.join(cur_dir, '__version__')
    if os.path.exists(version_file):
        return open(version_file).read().strip()

    return "development"


class App(Base):

    VERSION = get_version()

    def configure_routes(self):
        ctx.log.info("Configuring endpoints")

        endpoints = [
            {"prefix": "", "ctrl": gen_main_ctrl(self), "name": "main"},
        ]

        for ep in endpoints:
            ctx.log.debug("Controller[%s] @ %s/", ep["name"], ep["prefix"])
            self.flask.register_blueprint(ep["ctrl"], url_prefix=ep["prefix"])


def force_init_app():
    _ = app.__doc__  # Doing anything with the app triggers Proxy and creates the real object


app = Proxy(App)


import os
from flask import Blueprint, g, session, request
from uengine.errors import AuthenticationError
from uengine.context import ctx
from {{ project_name }}.models import Token, User


class AuthController(Blueprint):
    def __init__(self, *args, **kwargs):
        self.require_auth = kwargs.get("require_auth") or False
        if "require_auth" in kwargs:
            del kwargs["require_auth"]
        Blueprint.__init__(self, *args, **kwargs)
        self.before_request(self.set_current_user)

    @staticmethod
    def _get_user_from_authorization_header():
        if "Authorization" in request.headers:
            auth = request.headers["Authorization"].split()
            if len(auth) == 2 and auth[0] == "Token":
                token = Token.cache_get(auth[1])
                if token is not None and not token.expired:
                    token.prolongate()
                    return token.user
        return None

    @staticmethod
    def _get_user_from_session():
        user_id = session.get("user_id")
        if user_id:
            user = User.cache_get(user_id)
            return user
        return None

    @staticmethod
    def _get_user_from_x_api_auth_token():
        if "X-Api-Auth-Token" in request.headers:
            token = Token.cache_get(request.headers["X-Api-Auth-Token"])
            if token is not None and not token.expired:
                token.prolongate()
                return token.user

        return None

    def set_current_user(self):
        if ctx.envtype == 'development':
            username = os.getenv("DEV_USER")
            if username is not None:
                g.user = User.get(username)
                if g is not None:
                    return
        g.user = self._get_user_from_authorization_header() or \
            self._get_user_from_x_api_auth_token() or \
            self._get_user_from_session()
        if self.require_auth:
            if g.user is None:
                raise AuthenticationError()


from .test_abstract_model import TestAbstractModel
from .test_sharded_model import TestShardedModel
from .test_storable_model import TestStorableModel
from .test_submodel import TestShardedSubmodel, TestStorableSubmodel
from .test_afterlife import TestAfterlife



import itertools

from bson.objectid import ObjectId
from uengine.models.submodel import StorableSubmodel, ShardedSubmodel
from uengine.errors import WrongSubmodel, MissingSubmodel, InputDataError, IntegrityError
from .mongo_mock import MongoMockTest


class _BaseTestSubmodel(MongoMockTest):

    CLASS = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        class TestBaseModel(cls.CLASS):  # pylint: disable=inherit-non-class
            FIELDS = [
                "field1",
                "field2",
            ]
            COLLECTION = "test"

        class Submodel1(TestBaseModel):
            SUBMODEL = "submodel1"

        class Submodel2(TestBaseModel):
            SUBMODEL = "submodel2"

        TestBaseModel.register_submodel(Submodel1.SUBMODEL, Submodel1)
        TestBaseModel.register_submodel(Submodel2.SUBMODEL, Submodel2)

        cls.base_model = TestBaseModel
        cls.submodel1 = Submodel1
        cls.submodel2 = Submodel2

    def test_wrong_input(self):
        with self.assertRaises(WrongSubmodel):
            self.submodel1(_id=ObjectId(), field1="value", submodel="wrong")
        with self.assertRaises(MissingSubmodel):
            self.submodel1(_id=ObjectId(), field1="value")
        with self.assertRaises(InputDataError):
            self.submodel1(field1="value", submodel="my_submodel")
        with self.assertRaises(WrongSubmodel):
            obj = self.submodel1(field1="value")
            obj.submodel = "wrong"
            obj.save()

    def test_submodel_field(self):
        obj = self.submodel1()
        self.assertTrue(hasattr(obj, "submodel"))
        self.assertEqual(obj.submodel, self.submodel1.SUBMODEL)
        obj.save()
        obj.reload()
        self.assertEqual(obj.submodel, self.submodel1.SUBMODEL)
        db_obj = self.submodel1.get(obj._id)
        self.assertEqual(db_obj.submodel, self.submodel1.SUBMODEL)

    def test_inheritance(self):
        class Submodel1(self.base_model):
            SUBMODEL = "submodel1"

        class Submodel1_1(Submodel1):
            pass

        self.assertEqual(self.base_model.collection, Submodel1.collection)
        self.assertEqual(Submodel1.collection, Submodel1_1.collection)
        self.assertEqual(Submodel1.SUBMODEL, Submodel1_1.SUBMODEL)

    def test_abstract(self):
        with self.assertRaises(IntegrityError):
            self.base_model()

        with self.assertRaises(IntegrityError):
            class C(self.base_model):
                pass  # no SUBMODEL
            C()

        with self.assertRaises(IntegrityError):
            class C(self.submodel1):
                SUBMODEL = "c"
            self.submodel1.register_submodel("c", C)

    def _create_objs(self):
        """Returns two lists of objects. Objects in the same positions only differ in their submodel"""
        values = [1, 2, 3]
        objs1 = [self.submodel1(field1=v, field2=v) for v in values]
        objs2 = [self.submodel2(field1=v, field2=v) for v in values]
        for obj in itertools.chain(objs1, objs2):
            obj.save()

        return objs1, objs2

    def test_isolation_find(self):
        objs1, objs2 = self._create_objs()
        self.assertCountEqual(
            self.submodel1.find().all(),
            objs1,
        )
        self.assertCountEqual(
            self.submodel2.find().all(),
            objs2,
        )
        self.assertCountEqual(
            self.base_model.find().all(),
            objs1 + objs2,
        )

        self.assertCountEqual(
            self.submodel1.find({"field1": objs1[0].field1}).all(),
            [objs1[0]],
        )
        self.assertCountEqual(
            self.base_model.find({"field1": objs1[0].field1}).all(),
            [objs1[0], objs2[0]],
        )

    def test_isolation_update(self):
        objs1, objs2 = self._create_objs()
        obj1 = objs2[0]
        obj1.field2 = "new_value"
        self.submodel2.update_many(
            {"field1": obj1.field1},
            {"$set": {"field2": obj1.field2}}
        )
        self.assertCountEqual(
            self.base_model.find({"field2": obj1.field2}),
            [obj1]
        )

        obj1 = objs1[1]
        obj2 = objs2[1]
        obj1.field2 = "newer_value"
        obj2.field2 = "newer_value"
        self.base_model.update_many(
            {"field1": obj1.field1},
            {"$set": {"field2": obj1.field2}}
        )
        self.assertCountEqual(
            self.base_model.find({"field2": obj1.field2}),
            [obj1, obj2]
        )

    def test_isolation_destroy(self):
        objs1, objs2 = self._create_objs()
        to_destroy = objs2.pop()
        to_keep = objs1[-1]
        self.submodel2.destroy_many({"field1": to_destroy.field1})
        self.assertListEqual(
            self.submodel2.find({
                "field1": to_destroy.field1
            }).all(),
            []
        )
        self.assertListEqual(
            self.base_model.find({
                "field1": to_destroy.field1
            }).all(),
            [to_keep]
        )

        to_destroy = objs1[0]
        objs1 = objs1[1:]
        objs2 = objs2[1:]
        self.base_model.destroy_many({"field1": to_destroy.field1})
        self.assertCountEqual(
            self.base_model.find().all(),
            objs1 + objs2,
        )

    def test_double_register(self):
        with self.assertRaises(IntegrityError):
            class NewSubmodel(self.base_model):
                SUBMODEL = self.submodel2.SUBMODEL
            self.base_model.register_submodel(
                NewSubmodel.SUBMODEL, NewSubmodel)


class TestShardedSubmodel(_BaseTestSubmodel):

    # pylint: disable=arguments-differ
    class CLASS(ShardedSubmodel):

        def __init__(self, **data):
            data.setdefault("shard_id", "s1")
            super().__init__(**data)

        @classmethod
        def get(cls, *args, **kwargs):
            args = list(args)
            if len(args) == 1:
                args.insert(0, "s1")
            if not args:
                kwargs.setdefault("shard_id", "s1")
            return super().get(*args, **kwargs)

        @classmethod
        def find(cls, *args, **kwargs):
            args = list(args)
            if len(args) == 1:
                args.insert(0, "s1")
            if not args:
                kwargs.setdefault("shard_id", "s1")
            return super().find(*args, **kwargs)

        @classmethod
        def destroy_all(cls, *args, **kwargs):
            args = list(args)
            if not args:
                kwargs.setdefault("shard_id", "s1")
            return super().destroy_all(*args, **kwargs)

        @classmethod
        def destroy_many(cls, *args, **kwargs):
            args = list(args)
            if len(args) == 1:
                args.insert(0, "s1")
            if not args:
                kwargs.setdefault("shard_id", "s1")
            return super().destroy_many(*args, **kwargs)

        @classmethod
        def update_many(cls, *args, **kwargs):
            args = list(args)
            if len(args) == 2:
                args.insert(0, "s1")
            if not args:
                kwargs.setdefault("shard_id", "s1")
            return super().update_many(*args, **kwargs)


class TestStorableSubmodel(_BaseTestSubmodel):

    CLASS = StorableSubmodel



"""
- Make the base class first. Set COLLECTION explicitly - it is not
  generated automatically from the class name but inherited instead.
- Subclass your base class to define submodels. Set SUBMODEL to a string
  that will identify your submodel in the DB.
- You can further subclass your submodels. To avoid saving such abstract
  intermediate models do not set SUBMODEL.
- Register the submodel with the base model

It is possible to register an arbitrary function instead of a proper class.
It may be particularly useful if the correct class depends on something
other than `submodel` field. The function will get **data from the DB and
should return an model object.

If you decide to do it you will likely have to override _preprocess_query()
on your submodels to keep the expected find/destroy/update behaviour

>>> from uengine.models.submodel import StorableSubmodel

>>> class Animal(StorableSubmodel):
>>>     COLLECTION = "animal"
>>>     FIELDS = [
>>>         "weight",
>>>     ]

>>> class Fish(Animal):
>>>     SUBMODEL = "fish"
>>>     FIELDS = [
>>>         "scale_type",
>>>     ]

>>> class Mammal(Animal):
>>>     FIELDS = [
>>>         "penis_len",
>>>     ]

>>> class AquaticMammal(Mammal):
>>>     SUBMODEL = "aquatic_mammal"
>>>     FIELDS = [
>>>         "num_fins",
>>>     ]

>>>> class TerrestrialMammal(Mammal):
>>>     SUBMODEL = "terrestrial_mammal"
>>>     FIELDS = [
>>>         "num_legs",
>>>     ]

>>> Animal.register_submodel(Fish.SUBMODEL, Fish)
>>> Animal.register_submodel(TerrestrialMammal.SUBMODEL, TerrestrialMammal)
>>> Animal.register_submodel(AquaticMammal.SUBMODEL, AquaticMammal)
"""

from .abstract_model import ModelMeta
from .storable_model import StorableModel
from .sharded_model import ShardedModel
from uengine.errors import MissingSubmodel, UnknownSubmodel, WrongSubmodel, InputDataError, IntegrityError


class SubmodelMeta(ModelMeta):

    @staticmethod
    def _get_collection(model_cls, name, bases, dct):
        # SubModels inherit their collections from parent classes.
        if hasattr(model_cls, "COLLECTION") and model_cls.COLLECTION:  # Either set explicitly or inherited
            return model_cls.COLLECTION

        return None  # No autogeneration


class BaseSubmodelMixin:
    """Do not use this mixin directly. Subclass StorableSubmodel or
    ShardedSubmodel instead"""

    SUBMODEL = None
    AUXILIARY_SLOTS = [
        "SUBMODEL",
    ]
    FIELDS = [
        "submodel",
    ]
    REQUIRED_FIELDS = [
        "submodel",
    ]
    __submodel_loaders = None

    def __init__(self, **data):
        super().__init__(**data)
        if self.is_new:
            if not self.SUBMODEL:
                raise IntegrityError(f"Attempted to create an object of abstract model {self.__class__.__name__}")
            if "submodel" in data:
                raise InputDataError("Attempt to override submodel for a new object")
            self.submodel = self.SUBMODEL
        else:
            if not self.submodel:
                raise MissingSubmodel(f"{self.__class__.__name__} has no submodel in the DB. Bug?")
            self._check_submodel()

    def _check_submodel(self):
        if self.submodel != self.SUBMODEL:
            raise WrongSubmodel(
                f"Attempted to load {self.submodel} as {self.__class__.__name__}. Correct submodel "
                f"would be {self.SUBMODEL}. Bug?"
            )

    def _validate(self):
        super()._validate()
        self._check_submodel()

    @classmethod
    def register_submodel(cls, name, constructor):
        if cls.SUBMODEL:
            raise IntegrityError("Attempted to register a submodel with another submodel")
        if not cls.__submodel_loaders:
            cls.__submodel_loaders = {}
        if name in cls.__submodel_loaders:
            raise IntegrityError(f"Submodel {name} is already registered")
        cls.__submodel_loaders[name] = constructor

    @classmethod
    def from_data(cls, **data):
        if "submodel" not in data:
            raise MissingSubmodel(f"{cls.__name__} has no submodel in the DB. Bug?")
        if not cls.__submodel_loaders:
            return cls(**data)
        submodel_name = data["submodel"]
        if submodel_name not in cls.__submodel_loaders:
            raise UnknownSubmodel(f"Submodel {submodel_name} is not registered with {cls.__name__}")
        return cls.__submodel_loaders[submodel_name](**data)

    @classmethod
    def _preprocess_query(cls, query):
        if not cls.SUBMODEL:
            return query
        processed = {"submodel": cls.SUBMODEL}
        processed.update(query)
        return processed


class StorableSubmodel(BaseSubmodelMixin, StorableModel, metaclass=SubmodelMeta):
    pass


class ShardedSubmodel(BaseSubmodelMixin, ShardedModel, metaclass=SubmodelMeta):
    pass



documents_per_page = 10
app_secret_key = "_some_secret_key_"

flask_settings = {
    "SESSION_COOKIE_NAME": "_{{ project_name }}_session_id"
}

session_expiration_time = 86400 * 7 * 2  # 2 weeks
token_expiration_time = 86400 * 7 * 2  # 2 weeks

# If token_auto_prolongation is True, the token lifetime is calculated since updated_at which is updated
# every time user is authorized with the token.
#
# Otherwise it's calculated since created_at, therefore tokens live exactly `token_expiration_time` seconds
token_auto_prolongation = True

pymongo_extra = {
    "serverSelectionTimeoutMS": 1100,
    "socketTimeoutMS": 1100,
    "connectTimeoutMS": 1100,
}

database = {
    "meta": {
        "uri": "mongodb://localhost",
        "pymongo_extra": pymongo_extra,
        "dbname": "{{ project_name }}_dev",
    },
    "shards": {}
}

log_level = "debug"
log_format = "[%(asctime)s] %(levelname)s\t%(module)-8.8s:%(lineno)-3d %(request_id)-8s %(message)s"
debug = True



from uengine.utils import now
from uengine.models.storable_model import StorableModel


class User(StorableModel):

    FIELDS = (
        "_id",
        "ext_id",
        "username",
        "first_name",
        "last_name",
        "email",
        "avatar_url",
        "created_at",
        "updated_at",
        "supervisor",
    )

    KEY_FIELD = "username"

    DEFAULTS = {
        "first_name": "",
        "last_name": "",
        "avatar_url": "",
        "supervisor": False,
        "email": "",
        "ext_id": None,
        "created_at": now,
        "updated_at": now
    }

    REJECTED_FIELDS = (
        "password_hash",
        "supervisor",
        "created_at",
        "updated_at",
    )

    REQUIRED_FIELDS = (
        "username",
    )

    INDEXES = (
        ["username", {"unique": True}],
        "ext_id",
    )

    __slots__ = list(FIELDS)

    def touch(self):
        self.updated_at = now()

    def _before_save(self):
        self.touch()

    def get_auth_token(self):
        from .token import Token
        tokens = Token.find({"type": "auth", "user_id": self._id})
        for token in tokens:
            if not token.expired:
                return token
        token = Token(type="auth", user_id=self._id)
        token.save()

        return token




import mongomock

from unittest import TestCase
from uengine import ctx
from uengine.db import DB


class MongoMockTest(TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        try:
            del ctx.cfg
        except AttributeError:
            pass

        ctx.cfg = {
            "database": {
                "meta": {
                    "uri": "mongodb://zwfbpggeih",
                    "dbname": "unittest_meta",
                },
                "shards": {
                    "s1": {
                        "uri": "mongodb://zwfbpggeih",
                        "dbname": "unittest_s1",
                    },
                    "s2": {
                        "uri": "mongodb://zwfbpggeih",
                        "dbname": "unittest_s2",
                    },
                }
            }
        }

    def setUp(self) -> None:
        super().setUp()
        self.mongo_patcher = mongomock.patch(servers=[("zwfbpggeih")])
        self.mongo_patcher.start()
        try:
            del ctx.db
        except AttributeError:
            pass
        ctx.db = DB()

    def tearDown(self) -> None:
        try:
            del ctx.db
        except AttributeError:
            pass
        self.mongo_patcher.stop()
        super().tearDown()



import requests


class UEngineClientError(Exception):
    def __init__(self, message, status_code, path):
        super().__init__(message)
        self.message = message
        self.status_code = status_code
        self.path = path

    def __str__(self):
        return f"UEngineClientError status_code={self.status_code}, path={self.path} message:\n  {self.message}"

    def __repr__(self):
        return self.__str__()


class UEngineClient:

    RESTRICTED_OPTIONS = ('json', 'headers',)

    def __init__(self, baseurl, api_token, api_prefix=None, options=None):
        self.baseurl = baseurl
        self.api_prefix = api_prefix
        if self.api_prefix and self.api_prefix.endswith("/"):
            self.api_prefix = self.api_prefix[:-1]
        self.set_token(api_token)
        self.set_options(options)

    def set_options(self, options=None):
        self.options = options
        if not self.options:
            self.options = {}
        for opt in self.RESTRICTED_OPTIONS:
            if opt in self.options:
                del(self.options[opt])

    def set_token(self, api_token):
        self.api_token = api_token

    def request(self, method, path, json=None):
        if not path.startswith("/"):
            if self.api_prefix is None:
                raise ValueError(
                    "can't use path without leading slash when API prefix is not defined")
            path = f"{self.api_prefix}/{path}"

        url = f"{self.baseurl}{path}"
        headers = {"X-Api-Auth-Token": self.api_token}
        if method == "GET":
            response = requests.get(url, headers=headers, **self.options)
        else:
            response = requests.request(
                method, url, headers=headers, json=json, **self.options)

        resp_content = response.json()

        if response.status_code == 200:
            return resp_content

        err = resp_content.get("error") or resp_content
        raise UEngineClientError(err, response.status_code, path)

    def get(self, path):
        return self.request("GET", path)

    def delete(self, path):
        return self.request("DELETE", path)

    def post(self, path, json=None):
        return self.request("POST", path, json)

    def put(self, path, json=None):
        return self.request("PUT", path, json)

    def patch(self, path, json=None):
        return self.request("PATCH", path, json)



from uengine import ctx
from commands import Command
from datetime import datetime


class Sessions(Command):

    def init_argument_parser(self, parser):
        parser.add_argument("action", type=str, nargs=1,
                            choices=["cleanup", "count"])

    def run(self):

        action = self.args.action[0]
        if action == "count":
            total = ctx.db.meta.ro_conn["sessions"].find().count()
            expired = ctx.db.meta.ro_conn["sessions"].find(
                {"expiration": {"$lt": datetime.now()}}).count()
            print(f"Total number of sessions: {total}, expired: {expired}")
            if expired > 0:
                print("Use <micro.py sessions cleanup> to remove old sessions manually")

        elif action == "cleanup":
            print("Starting sessions clean up process...")
            count = ctx.db.meta.cleanup_sessions()
            if count == 0:
                print("There's no expired sessions to clean up")
            else:
                print(f"{count} expired sessions have been cleaned up")



from uengine.utils import uuid4_string, now
from uengine.models.storable_model import StorableModel
from uengine.context import ctx

DEFAULT_TOKEN_EXPIRATION_TIME = 87600 * 7 * 2
DEFAULT_TOKEN_AUTO_PROLONGATION = True


class Token(StorableModel):

    FIELDS = (
        "token",
        "type",
        "user_id",
        "created_at",
        "updated_at",
    )

    KEY_FIELD = "token"

    REQUIRED_FIELDS = (
        "type",
        "token",
        "user_id",
        "created_at",
        "updated_at",
    )

    DEFAULTS = {
        "type": "auth",
        "token": uuid4_string,
        "created_at": now,
        "updated_at": now,
    }

    INDEXES = (
        ["token", {"unique": True}],
        ["user_id", "type"]
    )

    # pylint: disable=attribute-defined-outside-init
    def touch(self):
        self.updated_at = now()

    def _before_save(self):
        self.touch()
        self.invalidate()

    def _before_delete(self):
        self.invalidate()

    def prolongate(self):
        auto_prolongation = ctx.cfg.get("token_auto_prolongation", DEFAULT_TOKEN_AUTO_PROLONGATION)
        if auto_prolongation:
            self.save()

    @property
    def user(self):
        from .user import User
        return User.cache_get(self.user_id)

    @property
    def expired(self):
        expiration_time = ctx.cfg.get("token_expiration_time", DEFAULT_TOKEN_EXPIRATION_TIME)
        if expiration_time <= 0:
            return False

        auto_prolongation = ctx.cfg.get("token_auto_prolongation", DEFAULT_TOKEN_AUTO_PROLONGATION)
        token_lifetime_start = self.updated_at if auto_prolongation else self.created_at
        token_lifetime = now() - token_lifetime_start

        return token_lifetime.total_seconds() > expiration_time


import flask
import unittest

from uengine import afterlife


class TestAfterlife(unittest.TestCase):

    def setUp(self):
        self.app = flask.Flask("afterlife_test_app")
        afterlife.Afterlife(self.app)
        self.client = self.app.test_client()

        @self.app.route("/")
        def root():
            return "Hello world"

        @self.app.route("/error")
        def error():
            flask.abort(500)

    def test_after_response(self):
        @self.app.after_response
        def after():
            after.called = True

        # Regular request
        after.called = False
        self.client.get("/")
        self.assertTrue(after.called)

        # HTTPException
        after.called = False
        self.client.get("/error")
        self.assertTrue(after.called)

        # Non-existent URL
        after.called = False
        self.client.get("/foobar")
        self.assertTrue(after.called)

        # The callback must be called only after the response is closed
        after.called = False
        response = self.client.get("/", buffered=False)
        response.get_data()
        self.assertFalse(after.called)
        response.close()
        self.assertTrue(after.called)

    def test_after_this_response(self):
        def callback():
            callback.called = True

        live_afterlife = True

        @self.app.route("/test")
        def test():
            if live_afterlife:
                afterlife.after_this_response(callback)
            return "Test"

        # Regular request
        callback.called = False
        live_afterlife = True
        self.client.get("/test")
        self.assertTrue(callback.called)

        # Request does not trigger after_this_response
        callback.called = False
        live_afterlife = False
        self.client.get("/test")
        self.assertFalse(callback.called)

        # Request without after_this_response
        callback.called = False
        live_afterlife = True
        self.client.get("/")
        self.assertFalse(callback.called)

        # The callback must be called only after the response is closed
        callback.called = False
        live_afterlife = True
        response = self.client.get("/test", buffered=False)
        response.get_data()
        self.assertFalse(callback.called)
        response.close()
        self.assertTrue(callback.called)

    def test_order(self):
        results = []

        # after_this_response executed first in the order they were registered
        @self.app.route("/test")
        def test():
            afterlife.after_this_response(lambda: results.append(1))
            afterlife.after_this_response(lambda: results.append(2))
            return "Test"

        # Then global after_response callbacks are executed in the order they were registered
        @self.app.after_response
        def after3():
            results.append(3)

        @self.app.after_response
        def after4():
            results.append(4)

        self.client.get("/test")
        self.assertListEqual(results, [1, 2, 3, 4])

    def test_errors(self):

        @self.app.after_response
        def after1():
            after1.called = True

        @self.app.after_response
        def after_fail():
            raise Exception("Oops, an expected exception happened")

        @self.app.after_response
        def after2():
            after2.called = True

        def callback1():
            callback1.called = True

        def callback_fail():
            raise Exception("Ooopsie, an expected exception happened")

        def callback2():
            callback2.called = True

        @self.app.route("/test")
        def test():
            afterlife.after_this_response(callback1)
            afterlife.after_this_response(callback_fail)
            afterlife.after_this_response(callback2)
            return "Test"

        self.client.get("/test")
        self.assertTrue(after1.called)
        self.assertTrue(after2.called)
        self.assertTrue(callback1.called)
        self.assertTrue(callback2.called)

    def test_context(self):
        self.assertFalse(afterlife.has_context())
        with self.assertRaises(RuntimeError):
            _ = afterlife.g.prop

        @self.app.after_response
        def callback():
            callback.prop_exists = "prop" in afterlife.g
            if callback.prop_exists:
                callback.prop_value = afterlife.g.prop

        @self.app.route("/test/<value>")
        def test(value):
            afterlife.g.prop = value
            return "Test"

        @self.app.route("/test2")
        def test2():
            return "Test2"

        self.client.get("/test/value1")
        self.assertTrue(callback.prop_exists)
        self.assertEqual(callback.prop_value, "value1")
        self.client.get("/test/value2")
        self.assertTrue(callback.prop_exists)
        self.assertEqual(callback.prop_value, "value2")

        # Context should reset every request
        self.client.get("/test2")
        self.assertFalse(callback.prop_exists)

    def test_no_context(self):
        def callback():
            callback.called = True

        afterlife.after_this_response(callback)
        self.assertTrue(callback.called)



import pymongo

from contextlib import contextmanager
from bson.objectid import ObjectId, InvalidId
from time import sleep
from datetime import datetime
from random import randint
from pymongo.errors import ServerSelectionTimeoutError
from uengine.errors import InvalidShardId

from . import ctx

MONGO_RETRIES = 6
MONGO_RETRIES_RO = 6
RETRY_SLEEP = 3  # 3 seconds


class AbortTransaction(Exception):
    pass


def intercept_mongo_errors_rw(func):
    def wrapper(*args, **kwargs):
        if "retries_left" in kwargs:
            retries_left = kwargs["retries_left"]
            del kwargs["retries_left"]
        else:
            retries_left = MONGO_RETRIES

        try:
            result = func(*args, **kwargs)
        except ServerSelectionTimeoutError:
            ctx.log.error(
                "ServerSelectionTimeout in db module for read/write operations")
            retries_left -= 1
            if retries_left == MONGO_RETRIES / 2:
                ctx.log.error(
                    "Mongo connection %d retries passed with no result, "
                    "trying to reinstall connection",
                    MONGO_RETRIES / 2
                )
                db_obj = args[0]
                db_obj.reset_conn()
            if retries_left == 0:
                ctx.log.error(
                    "Mongo connection %d retries more passed with no result, giving up", MONGO_RETRIES / 2)
                raise

            sleep(RETRY_SLEEP)
            kwargs["retries_left"] = retries_left
            return wrapper(*args, **kwargs)

        return result

    return wrapper


def intercept_mongo_errors_ro(func):
    def wrapper(*args, **kwargs):
        if "retries_left" in kwargs:
            retries_left = kwargs["retries_left"]
            del kwargs["retries_left"]
        else:
            retries_left = MONGO_RETRIES

        try:
            result = func(*args, **kwargs)
        except pymongo.errors.ServerSelectionTimeoutError:
            ctx.log.error(
                "ServerSelectionTimeout in db module for read-only operations")
            retries_left -= 1
            if retries_left == MONGO_RETRIES_RO / 2:
                ctx.log.error(
                    "Mongo readonly connection %d retries passed, switching "
                    "readonly operations to read-write socket",
                    MONGO_RETRIES_RO / 2
                )
                db_obj = args[0]
                db_obj._ro_conn = db_obj.conn  # pylint: disable=protected-access
            if retries_left == 0:
                raise

            sleep(RETRY_SLEEP)
            kwargs["retries_left"] = retries_left
            return wrapper(*args, **kwargs)

        return result

    return wrapper


class ObjectsCursor:

    def __init__(self, cursor, obj_class, shard_id=None):
        self.obj_class = obj_class
        self.cursor = cursor
        self._shard_id = shard_id

    def all(self):
        return list(self)

    def limit(self, *args, **kwargs):
        self.cursor.limit(*args, **kwargs)
        return self

    def skip(self, *args, **kwargs):
        self.cursor.skip(*args, **kwargs)
        return self

    def sort(self, *args, **kwargs):
        self.cursor.sort(*args, **kwargs)
        return self

    def __iter__(self):
        for item in self.cursor:
            if self._shard_id:
                item["shard_id"] = self._shard_id
            yield self.obj_class(**item)

    def __getitem__(self, item):
        attrs = self.cursor.__getitem__(item)
        if self._shard_id:
            attrs["shard_id"] = self._shard_id
        return self.obj_class(**attrs)

    def __getattr__(self, item):
        return getattr(self.cursor, item)


def pick_rw_shard_id():
    idx = randint(0, len(ctx.db.rw_shards)-1)
    return ctx.db.rw_shards[idx]


class _DB:

    @contextmanager
    def transaction(self):
        client = self.get_rw_client()
        try:
            self._session = client.start_session()
            self._session.start_transaction()
            yield self._session
            self._session.commit_transaction()
        except Exception as e:
            self._session.abort_transaction()
            if not isinstance(e, AbortTransaction):
                raise
        finally:
            self._session.end_session()
            self._session = None

    def __init__(self, dbconf, shard_id=None):
        self._config = dbconf
        self._rw_client = None
        self._ro_client = None
        self._conn = None
        self._ro_conn = None
        self._shard_id = shard_id
        self._session = None

    def reset_conn(self):
        self._rw_client = None
        self._conn = None

    def reset_ro_conn(self):
        self._ro_client = None
        self._ro_conn = None

    def get_rw_client(self):
        if not self._rw_client:
            client_kwargs = self._config.get("pymongo_extra", {})
            self._rw_client = pymongo.MongoClient(self._config["uri"], **client_kwargs)
        return self._rw_client

    def get_ro_client(self):
        if not self._ro_client:
            client_kwargs = self._config.get("pymongo_extra", {})
            if "uri_ro" in self._config:
                self._ro_client = pymongo.MongoClient(self._config["uri_ro"], **client_kwargs)
        return self._ro_client

    def init_ro_conn(self):
        ctx.log.info("Creating a read-only mongo connection")
        database = self._config.get('dbname')
        ro_client = self.get_ro_client()
        if ro_client:
            # AUTHENTICATION
            if 'username' in self._config and 'password' in self._config:
                username = self._config["username"]
                password = self._config['password']
                ro_client[database].authenticate(username, password)
            self._ro_conn = ro_client[database]
        else:
            ctx.log.info(
                "No uri_ro option found in configuration, falling back to read/write default connection")
            self._ro_conn = self.conn

    def init_conn(self):
        ctx.log.info("Creating a read/write mongo connection")
        client = self.get_rw_client()
        database = self._config['dbname']
        # AUTHENTICATION
        if 'username' in self._config and 'password' in self._config:
            username = self._config["username"]
            password = self._config['password']
            client[database].authenticate(username, password)
        self._conn = client[database]

    @property
    def conn(self):
        if self._conn is None:
            self.init_conn()
        return self._conn

    @property
    def ro_conn(self):
        if self._ro_conn is None:
            self.init_ro_conn()
        return self._ro_conn

    @intercept_mongo_errors_ro
    @ctx.line_profiler
    def get_obj(self, cls, collection, query):
        if not isinstance(query, dict):
            try:
                query = {'_id': ObjectId(query)}
            except InvalidId:
                pass
        data = self.ro_conn[collection].find_one(query, session=self._session)
        if data:
            if self._shard_id:
                data["shard_id"] = self._shard_id
            return cls(**data)

        return None

    @intercept_mongo_errors_ro
    def get_obj_id(self, collection, query):
        return self.ro_conn[collection].find_one(query, projection=(), session=self._session)['_id']

    @intercept_mongo_errors_ro
    def get_objs(self, cls, collection, query, **kwargs):
        if self._session:
            kwargs["session"] = self._session
        cursor = self.ro_conn[collection].find(query, **kwargs)
        return ObjectsCursor(cursor, cls, shard_id=self._shard_id)

    @intercept_mongo_errors_ro
    def get_objs_projected(self, collection, query, projection, **kwargs):
        if self._session:
            kwargs["session"] = self._session
        cursor = self.ro_conn[collection].find(
            query, projection=projection, **kwargs)
        return cursor

    @intercept_mongo_errors_ro
    def get_aggregated(self, collection, pipeline, **kwargs):
        if self._session:
            kwargs["session"] = self._session
        cursor = self.ro_conn[collection].aggregate(pipeline, **kwargs)
        return cursor

    def get_objs_by_field_in(self, cls, collection, field, values, **kwargs):
        return self.get_objs(
            cls,
            collection,
            {
                field: {
                    '$in': values,
                },
            },
            **kwargs
        )

    @intercept_mongo_errors_rw
    def save_obj(self, obj):
        if obj.is_new:
            # object to_dict() method should always return all fields
            data = obj.to_dict(include_restricted=True)
            # although with the new object we shouldn't pass _id=null to mongo
            del data["_id"]
            inserted_id = self.conn[obj.collection].insert_one(
                data, session=self._session).inserted_id
            obj._id = inserted_id
        else:
            self.conn[obj.collection].replace_one(
                {'_id': obj._id}, obj.to_dict(include_restricted=True), upsert=True, session=self._session)

    @intercept_mongo_errors_rw
    def delete_obj(self, obj):
        if obj.is_new:
            return
        self.conn[obj.collection].delete_one({'_id': obj._id}, session=self._session)

    @intercept_mongo_errors_rw
    def find_and_update_obj(self, obj, update, when=None):
        query = {"_id": obj._id}
        if when:
            assert "_id" not in when
            query.update(when)

        new_data = self.conn[obj.collection].find_one_and_update(
            query,
            update,
            return_document=pymongo.ReturnDocument.AFTER,
            session=self._session
        )
        if new_data and self._shard_id:
            new_data["shard_id"] = self._shard_id
        return new_data

    @intercept_mongo_errors_rw
    def delete_query(self, collection, query):
        return self.conn[collection].delete_many(query, session=self._session)

    @intercept_mongo_errors_rw
    def update_query(self, collection, query, update):
        return self.conn[collection].update_many(query, update, session=self._session)

    # SESSIONS

    @intercept_mongo_errors_ro
    def get_session(self, sid, collection='sessions'):
        return self.ro_conn[collection].find_one({'sid': sid})

    @intercept_mongo_errors_rw
    def update_session(self, sid, data, expiration, collection='sessions'):
        self.conn[collection].update(
            {'sid': sid}, {'sid': sid, 'data': data, 'expiration': expiration}, True)

    @intercept_mongo_errors_rw
    def cleanup_sessions(self, collection='sessions'):
        return self.conn[collection].remove({'expiration': {'$lt': datetime.now()}})["n"]


class DB:

    INFO_FIELDS = (
        "allocator",
        "bits",
        "debug",
        "gitVersion",
        "javascriptEngine",
        "maxBsonObjectSize",
        "modules",
        "ok",
        "openssl",
        "storageEngines",
        "sysInfo",
        "version",
        "versionArray"
    )

    def __init__(self):
        self.meta = _DB(ctx.cfg["database"]["meta"])
        self.shards = {}
        if "shards" in ctx.cfg["database"]:
            for shard_id, config in ctx.cfg["database"]["shards"].items():
                self.shards[shard_id] = _DB(config, shard_id)

        if "open_shards" in ctx.cfg["database"]:
            self.rw_shards = ctx.cfg["database"]["open_shards"]
        else:
            self.rw_shards = list(self.shards.keys())

    def get_shard(self, shard_id):
        if shard_id not in self.shards:
            raise InvalidShardId(f"shard {shard_id} doesn't exist")
        return self.shards[shard_id]

    def mongodb_info(self):

        def sys_info(raw_info):
            return {k: v for k, v in raw_info.items() if k in self.INFO_FIELDS}

        return dict(
            meta=sys_info(self.meta.conn.client.server_info()),
            shards={shard_id: sys_info(
                self.shards[shard_id].conn.client.server_info()) for shard_id in self.shards}
        )



from functools import partial
from uengine import ctx
from uengine.utils import resolve_id
from uengine.errors import NotFound, ModelDestroyed, IntegrityError
from uengine.cache import req_cache_get, req_cache_set, req_cache_has_key, req_cache_delete
from datetime import datetime
from bson.objectid import ObjectId

from .abstract_model import AbstractModel, save_required


class StorableModel(AbstractModel):

    def __init__(self, **kwargs):
        AbstractModel.__init__(self, **kwargs)

    @property
    def _db(self):
        if not self.collection:
            raise IntegrityError(
                f"There is no DB for abstract model: {self.__class__.__name__}")
        return ctx.db.meta

    def _save_to_db(self):
        self._db.save_obj(self)

    def update(self, data, skip_callback=False, invalidate_cache=True):
        for field in self.FIELDS:
            if field in data and field not in self.REJECTED_FIELDS and field != "_id":
                self.__setattr__(field, data[field])
        self.save(skip_callback=skip_callback,
                  invalidate_cache=invalidate_cache)

    @save_required
    def db_update(self, update, when=None, reload=True, invalidate_cache=True):
        """
        :param update: MongoDB update query
        :param when: filter query. No update will happen if it does not match
        :param reload: Load the new stat into the object (Caution: if you do not do this
                       the next save() will overwrite updated fields)
        :return: True if the document was updated. Otherwise - False
        """
        new_data = self._db.find_and_update_obj(self, update, when)
        if invalidate_cache and new_data:
            self.invalidate()

        if reload and new_data:
            tmp = self.from_data(**new_data)
            self._reload_from_obj(tmp)

        return bool(new_data)

    def _delete_from_db(self):
        self._db.delete_obj(self)

    def _refetch_from_db(self):
        return self.find_one({"_id": self._id})

    def reload(self):
        if self.is_new:
            return
        tmp = self._refetch_from_db()
        if tmp is None:
            raise ModelDestroyed("model has been deleted from db")
        for field in self.FIELDS:
            if field == "_id":
                continue
            value = getattr(tmp, field)
            setattr(self, field, value)

    @classmethod
    # E.g. override if you want model to always return a subset of documents in its collection
    def _preprocess_query(cls, query):
        return query

    @classmethod
    def find(cls, query=None, **kwargs):
        if not query:
            query = {}
        return ctx.db.meta.get_objs(cls.from_data, cls.collection, cls._preprocess_query(query), **kwargs)

    @classmethod
    def aggregate(cls, pipeline, query=None, **kwargs):
        if not query:
            query = {}
        pipeline = [{"$match": cls._preprocess_query(query)}] + pipeline
        return ctx.db.meta.get_aggregated(cls.collection, pipeline, **kwargs)

    @classmethod
    def find_projected(cls, query=None, projection=('_id',), **kwargs):
        if not query:
            query = {}
        return ctx.db.meta.get_objs_projected(cls.collection, cls._preprocess_query(query),
                                              projection=projection, **kwargs)

    @classmethod
    def find_one(cls, query, **kwargs):
        return ctx.db.meta.get_obj(cls.from_data, cls.collection, cls._preprocess_query(query), **kwargs)

    @classmethod
    def get(cls, expression, raise_if_none=None):
        if expression is None:
            return None

        expression = resolve_id(expression)
        if isinstance(expression, ObjectId):
            query = {"_id": expression}
        else:
            expression = str(expression)
            query = {cls.KEY_FIELD: expression}
        res = cls.find_one(query)
        if res is None and raise_if_none is not None:
            if isinstance(raise_if_none, Exception):
                raise raise_if_none
            else:
                raise NotFound(f"{cls.__name__} not found")
        return res

    @classmethod
    def _cache_get(cls, cache_key, getter, constructor=None):
        d1 = datetime.now()
        if not constructor:
            constructor = cls.from_data

        if req_cache_has_key(cache_key):
            data = req_cache_get(cache_key)
            td = (datetime.now() - d1).total_seconds()
            ctx.log.debug("ModelCache L1 HIT %s %.3f seconds", cache_key, td)
            return constructor(**data)

        if ctx.cache.has(cache_key):
            data = ctx.cache.get(cache_key)
            req_cache_set(cache_key, data)
            td = (datetime.now() - d1).total_seconds()
            ctx.log.debug("ModelCache L2 HIT %s %.3f seconds", cache_key, td)
            return constructor(**data)

        obj = getter()
        if obj:

            data = obj.to_dict()
            ctx.cache.set(cache_key, data)
            req_cache_set(cache_key, data)

        td = (datetime.now() - d1).total_seconds()
        ctx.log.debug("ModelCache MISS %s %.3f seconds", cache_key, td)
        return obj

    @classmethod
    def cache_get(cls, expression, raise_if_none=None):
        if expression is None:
            return None
        cache_key = f"{cls.collection}.{expression}"
        getter = partial(cls.get, expression, raise_if_none)
        return cls._cache_get(cache_key, getter)

    @staticmethod
    def _invalidate(cache_key_id, cache_key_keyfield=None):
        ctx.log.debug("ModelCache DELETE %s", cache_key_id)
        cr_layer1_id = req_cache_delete(cache_key_id)
        cr_layer2_id = ctx.cache.delete(cache_key_id)
        cr_layer1_keyfield = None
        cr_layer2_keyfield = None
        if cache_key_keyfield:
            ctx.log.debug("ModelCache DELETE %s", cache_key_keyfield)
            cr_layer1_keyfield = req_cache_delete(cache_key_keyfield)
            cr_layer2_keyfield = ctx.cache.delete(cache_key_keyfield)

        return cr_layer1_id, cr_layer1_keyfield, cr_layer2_id, cr_layer2_keyfield

    def invalidate(self):
        cache_key_id = f"{self.collection}.{self._id}"
        cache_key_keyfield = None
        if self.KEY_FIELD is not None and self.KEY_FIELD != "_id":
            cache_key_keyfield = f"{self.collection}.{getattr(self, self.KEY_FIELD)}"
        return self._invalidate(cache_key_id, cache_key_keyfield)

    @classmethod
    def destroy_all(cls):
        ctx.db.meta.delete_query(cls.collection, cls._preprocess_query({}))

    @classmethod
    def destroy_many(cls, query):
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        ctx.db.meta.delete_query(cls.collection, cls._preprocess_query(query))

    @classmethod
    def update_many(cls, query, attrs):
        # warning: being a faster method than traditional model manipulation,
        # this method doesn't provide any lifecycle callback for independent
        # objects
        ctx.db.meta.update_query(
            cls.collection, cls._preprocess_query(query), attrs)



"""
>>> import time
>>> import flask
>>> import uuid
>>> from uengine.afterlife import Afterlife, after_this_response, g

>>> app = flask.Flask("after_response")
>>> Afterlife(app)

>>> @app.after_response
>>> def say_oops():
>>>     time.sleep(1)
>>>     print("Oops...")
>>>     raise RuntimeError("Wheee!")

>>> @app.after_response
>>> def say_bye():
>>>     time.sleep(0.1)
>>>     print(f"Bye {g.request_id}")

>>> @app.route("/")
>>> def home():
>>>     g.request_id = str(uuid.uuid4())
>>>
>>>     @after_this_response
>>>     def say_wapapapapapapow():
>>>         time.sleep(1.5)
>>>         print("Wa-pa-pa-pa-pa-pa-pow")
>>>         time.sleep(0.01)
>>>
>>>     after_this_response(lambda: print("What does the fox say?"))
>>>
>>>     return "Success!\n"
"""

import functools

from itertools import chain
from werkzeug.local import LocalStack, LocalProxy
from werkzeug.wsgi import ClosingIterator


class Afterlife:
    def __init__(self, app=None, patch_test_client=True, logger=None):
        """
        :param app: Flask app to apply middleware to
        :param patch_test_client: Make test client buffer responses by default.
                                  It is makes the client close the request before
                                  returning, thus running all afterlife callbacks.
                                  See Werkzeug.test.Client.open() for details.
        :param logger: Logger to use. By default app.logger will be used
        """
        self.after_response_functions = []
        self.patch_test_client = patch_test_client
        self.logger = logger
        if app:
            self.init_app(app)

    def after_response(self, func):
        self.after_response_functions.append(func)
        return func

    def init_app(self, app):
        # install extension
        app.after_response = self.after_response
        # install middleware
        app.wsgi_app = _AfterlifeMiddleware(app.wsgi_app, self.run_after_response, app.logger)
        if not self.logger:
            self.logger = app.logger

        if not self.patch_test_client:
            return

        test_client_orig = app.test_client

        @functools.wraps(app.test_client)
        def patched_test_client(*args, **kwargs):
            client = test_client_orig(*args, **kwargs)
            client.open = functools.partial(client.open, buffered=True)
            return client

        app.test_client = patched_test_client

    def run_after_response(self):
        funcs = chain(_after_response_functions, self.after_response_functions)
        for func in funcs:
            try:
                func()
            except Exception:
                self.logger.exception("Unhandled exception")


def after_this_response(func):
    """
    Schedules a callable to be executed after current request has been responded to
    If called outside of any request executes func immediately synchronously.
    :param func: Callable to schedule after this response
    :return func itself. Can be used as a decorator
    """
    if has_context():
        _after_response_functions.append(func)
    else:
        func()
    return func


class _AfterlifeMiddleware:
    def __init__(self, application, after_response_callback, logger):
        self.application = application
        self.after_response_callback = after_response_callback
        self.logger = logger

    @staticmethod
    def set_local_store():
        _afterlife_ctx_stack.push(AfterlifeContext())

    @staticmethod
    def clear_local_store():
        _afterlife_ctx_stack.pop()

    def _after_response(self):
        try:
            self.after_response_callback()
        finally:
            self.clear_local_store()

    def __call__(self, environ, start_response):
        self.set_local_store()
        iterator = self.application(environ, start_response)
        return ClosingIterator(iterator, [self._after_response])


class AfterlifeContext:

    def __init__(self):
        self.g = _AfterlifeGlobals()
        self.after_response_functions = []


_not_set = object()


class _AfterlifeGlobals:

    def get(self, name, default=None):
        return self.__dict__.get(name, default)

    def pop(self, name, default=_not_set):
        if default is _not_set:
            return self.__dict__.pop(name)
        else:
            return self.__dict__.pop(name, default)

    def setdefault(self, name, default=None):
        return self.__dict__.setdefault(name, default)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)


def _from_ctx(attr):
    top = _afterlife_ctx_stack.top
    if top is None:
        raise RuntimeError("Working outside of afterlife request context")
    return getattr(top, attr)


def has_context():
    return _afterlife_ctx_stack.top is not None


_afterlife_ctx_stack = LocalStack()
_after_response_functions = LocalProxy(lambda: _from_ctx("after_response_functions"))
g = LocalProxy(lambda: _from_ctx("g"))



import os
from cachelib import SimpleCache
from uengine.context import ctx
from {{ project_name }} import app, force_init_app

force_init_app()

if isinstance(ctx.cache, SimpleCache):
    if not os.getenv("UENGINE_USE_SIMPLE_CACHE"):
        raise RuntimeError("""Running your application under uwsgi control without centralized cache
is dangerous. If you rely on cache invalidation your app instances will be
able to invalidate their own in-memory cache while others will consider the
same cache keys still valid.

If you're sure it's OK for your project please set:
 
UENGINE_USE_SIMPLE_CACHE=1

""")

app_callable = app.flask




class AbstractQueue:

    ACK_POSTFIX = ":ack"

    def __init__(self, qcfg):
        self.cfg = qcfg

    def enqueue(self, task):
        raise NotImplementedError("abstract queue")

    def ack(self, task_id):
        raise NotImplementedError("abstract queue")

    def subscribe(self):
        raise NotImplementedError("abstract queue")

    def get_message(self, **kwrds):
        raise NotImplementedError("abstract queue")

    @property
    def tasks(self):
        raise NotImplementedError("abstract queue")



from commands import Command
from {{project_name}} import app
from uengine import ctx
from uengine.utils import get_modules
import importlib
import os.path


class Index(Command):

    def init_argument_parser(self, parser):
        parser.add_argument("-w", "--overwrite", dest="overwrite", action="store_true", default=False,
                            help="Overwrite existing indexes in case of conflicts")

    def run(self):

        ctx.log.info("Creating indexes")
        models_directory = os.path.join(
            app.base_dir, "{{ project_name }}/models")
        modules = [x for x in get_modules(models_directory) if x not in
                   ("storable_model", "abstract_model", "sharded_model")]
        for mname in modules:
            module = importlib.import_module(
                "{{ project_name }}.models.%s" % mname)
            for attr in dir(module):
                if attr.startswith("__") or attr in ("StorableModel", "AbstractModel", "ShardedModel"):
                    continue
                obj = getattr(module, attr)
                if hasattr(obj, "ensure_indexes"):
                    ctx.log.info(
                        "Creating indexes for %s, collection %s", attr, obj.collection)
                    obj.ensure_indexes(True, self.args.overwrite)
        ctx.log.info("Creating sessions indexes")
        ctx.db.meta.conn["sessions"].create_index(
            "sid", unique=True, sparse=False)



import pylint.lint

from commands import Command


class Lint(Command):

    NO_ARGPARSE = True

    def run(self):
        lint_args = [
            "exyaru",
            "commands",
            "uengine",
            "micro",
            "wsgi",
        ]
        lint_args += self.raw_args
        linter = pylint.lint.Run(lint_args, do_exit=False)
        return linter.linter.msg_status



#!/usr/bin/env python3
import argparse
import os
from jinja2 import Environment, FileSystemLoader

def create_project(args):
    project_name = args.name[0]

    if args.inplace:
        root_dir = "."
    else:
        root_dir = project_name
        if os.path.isdir(project_name):
            raise RuntimeError(f"{project_name} exists")
        os.mkdir(root_dir)

    template_path = os.path.join(os.path.dirname(__file__), "project_source_templates")
    env = Environment(loader=FileSystemLoader(template_path))

    for folder, _, files in os.walk(template_path):
        if "__pycache__" in folder:
            continue
        folder = folder[len(template_path)+1:]
        dst_dir = folder.replace("__appdir__", project_name)
        dst_dir = os.path.join(root_dir, dst_dir)
        if not os.path.isdir(dst_dir):
            os.mkdir(dst_dir)
        for f in files:
            if f.endswith(".pyc"):
                continue
            dst_file = os.path.join(dst_dir, f)
            tmpl = env.get_template(os.path.join(folder, f))
            with open(dst_file, "w") as outf:
                outf.write(tmpl.render(project_name=project_name, auth=True))
    micro_binary = os.path.join(root_dir, "micro.py")
    os.chmod(micro_binary, 0o755)

def main():
    parser = argparse.ArgumentParser(prog="uengine")
    subparsers = parser.add_subparsers(help="sub-command help", dest="action")

    create_parser = subparsers.add_parser("create", help="create a new uengine-based project")
    create_parser.add_argument("name", nargs=1, help="name of your project")
    create_parser.add_argument("-i", "--inplace", dest="inplace", action="store_true", default=False, help="create project structure in current directory instead of [project_name]/")

    args = parser.parse_args()
    if args.action is None:
        parser.print_usage()
        return

    if args.action == "create":
        create_project(args)


if __name__ == '__main__':
    main()


from commands import Command
from commands.send import MessageTask, BaseTask
from uengine import ctx


class Tasks(Command):

    def run(self):
        BaseTask.register(MessageTask)
        ctx.log.info("running task list server")
        q = ctx.queue
        q.subscribe()
        for task in q.tasks:
            ctx.log.info("Got task %s", task)



from commands import Command
from uengine import ctx
from uengine.queue import BaseTask
from sys import stdin


class MessageTask(BaseTask):
    TYPE = "MESSAGE"


class Send(Command):

    def run(self):
        ctx.log.info("starting up task client")
        print("Type in messages and send them with Enter:")
        BaseTask.register(MessageTask)

        while True:
            message = stdin.readline().strip()
            if not message:
                break
            t = MessageTask({"message": message})
            ctx.log.info("sending task %s", t)
            ctx.queue.enqueue(t)
            ctx.log.info("sent task %s", t)

        ctx.log.info("shutting down task client")


import os
import pickle
import functools
from hashlib import sha256

from uengine import ctx
from uengine.utils import now
from datetime import timedelta
from .cache import DEFAULT_CACHE_PREFIX, DEFAULT_CACHE_TIMEOUT, _get_cache_key


class FileCache:

    def __init__(self, cache_dir=""):
        self.cache_dir = cache_dir
        if not os.path.isdir(cache_dir):
            try:
                os.makedirs(cache_dir, 0o755)
            except (OSError, IOError) as e:
                ctx.log.error(
                    "error creating cache directory %s: %s", cache_dir, e)
                self.initialized = False
                return
        self.initialized = True

    def __getpath(self, key):
        hashed_key = sha256(key.encode())
        return os.path.join(self.cache_dir, hashed_key.hexdigest())

    def set(self, key, value, timeout=None):
        if not self.initialized:
            return False
        path = self.__getpath(key)
        try:
            with open(path, "wb") as cf:
                expires = None
                if timeout is not None:
                    td = timedelta(seconds=timeout)
                    expires = now() + td
                data = {
                    "value": value,
                    "expires": expires
                }
                pickle.dump(data, cf)
        except Exception as e:
            ctx.log.error("error writing filecache %s: %s", path, e)
            return False
        return True

    def __load(self, key):
        if not self.initialized:
            return None
        path = self.__getpath(key)
        if not os.path.isfile(path):
            return None
        try:
            with open(path, "rb") as cf:
                data = pickle.load(cf)
                return data
        except Exception as e:
            ctx.log.error("error loading filecache %s: %s", path, e)
            return None

    def get(self, key):
        data = self.__load(key)
        if data is None:
            return None
        if data["expires"] is not None and now() > data["expires"]:
            self.delete(key)
            return None
        return data["value"]

    def has(self, key):
        data = self.__load(key)
        if data is None:
            return False
        if data["expires"] is not None and now() > data["expires"]:
            self.delete(key)
            return False
        return True

    def expires(self, key):
        data = self.__load(key)
        if data is None:
            return None
        return data["expires"]

    def delete(self, key):
        if not self.initialized:
            return False
        path = self.__getpath(key)
        if os.path.isfile(path):
            try:
                os.unlink(path)
                return True
            except Exception as e:
                ctx.log.error("error deleting filecache %s: %s", path, e)
        return False


def file_cached_function(cache_key_prefix=DEFAULT_CACHE_PREFIX, cache_timeout=DEFAULT_CACHE_TIMEOUT, positive_only=False):
    def cache_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            cache_key, _ = _get_cache_key(
                cache_key_prefix, func.__name__, args, kwargs)
            t1 = now()
            if ctx.filecache.has(cache_key):
                value = ctx.filecache.get(cache_key)
                ctx.log.debug("FileCache HIT %s (%.3f seconds)",
                              cache_key, (now() - t1).total_seconds())
            else:
                value = func(*args, **kwargs)
                if value or not positive_only:
                    ctx.filecache.set(cache_key, value, timeout=cache_timeout)
                ctx.log.debug("FileCache MISS %s (%.3f seconds)",
                              cache_key, (now() - t1).total_seconds())
            return value
        return wrapper
    return cache_decorator



from {{project_name}}.models import User, Token
import logging
import mongomock

from commands import Command, _all_tests
from unittest import main
from uengine import ctx
from {{project_name}} import force_init_app
from {{project_name}} import tests


class Test(Command):

    NO_ARGPARSE = True

    @mongomock.patch()
    def run(self):
        force_init_app()
        ctx.log.setLevel(logging.ERROR)
        argv = ['micro.py test', '--buffer'] + self.raw_args
        test_program = main(argv=argv, module=_all_tests, exit=False)
        if test_program.result.wasSuccessful():
            return 0

        return 1



from setuptools import setup, find_packages


setup(
    name="uengine",
    version="3.5.3",
    description="a micro webframework based on flask and pymongo",
    url="https://github.com/viert/uengine",
    author="Pavel Vorobyov",
    author_email="aquavitale@yandex.ru",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "uengine": ["*.txt", "*.py"]
    },
    install_requires=[
        "jinja2",
        "flask",
        "pymongo",
        "cachelib",
        "lazy_object_proxy",
        "ipython",
        "pylint",
        "mongomock",
        "requests",
    ],
    entry_points={
        "console_scripts": [
            "uengine=uengine.__main__:main",
        ]
    }
)



import flask

from uengine import ctx
from uengine.cache import check_cache
from uengine.api import json_response
from .auth_controller import AuthController


def gen_main_ctrl(app):
    main_ctrl = AuthController("main", __name__, require_auth=False)

    def index():
        routes = []
        for rule in app.flask.url_map.iter_rules():
            routes.append({
                "endpoint": rule.endpoint,
                "route": rule.rule,
                "methods": rule.methods
            })
        return json_response({"routes": routes})

    def app_info():
        results = dict(
            app={"name": "testapp", "version": app.version},
            mongodb=ctx.db.mongodb_info(),
            cache={"type": ctx.cache.__class__.__name__, "active": check_cache()},
            flask_version=flask.__version__
        )

        for shard_id, shard in ctx.db.shards.items():
            results["mongodb"]["shards"][shard_id] = shard.conn.client.server_info()

        return json_response({"app_info": results})

    main_ctrl.route("/")(index)
    main_ctrl.route("/app_info")(app_info)

    return main_ctrl


import sys
import pkgutil
import inspect
from argparse import ArgumentParser, REMAINDER


class Command:
    """Base class for CLI commands
    You should override at least run() method
        run() - method is being called to do the main job
        init_argument_parser(parser) - override this method if you want to accept arguments
        NAME - if this property is not None it is used as command name.
               Otherwise command name is generated from class name
        DESCRIPTION - description that is used in help messages. Consider setting it to something meaningful.
    """
    NAME = None
    DESCRIPTION = None
    NO_ARGPARSE = False

    def __init__(self):
        if self.NAME is None:
            self.NAME = self.__class__.__name__.lower()
        if self.DESCRIPTION is None:
            self.DESCRIPTION = '"%s" has no DESCRIPTION' % (self.NAME,)
        self.args = None
        self.raw_args = None

    def init_argument_parser(self, parser):
        """
        This method is called to configure argument subparser for command
        Override it if you need to accept arguments
          - parser: argparse.ArgumentParser to fill with arguments
        """

    def run(self):
        """
        This is a main method of command. Override it and do all the job here
        Command arguments can be read from self.args
        The return value from this method will be used as CLI exit code
        """
        raise NotImplementedError()


def is_a_command_class(obj):
    return inspect.isclass(obj) and Command in inspect.getmro(obj) and obj != Command


def load_commands_from_module(module):
    return [obj for _, obj in inspect.getmembers(module) if is_a_command_class(obj)]


def load_commands_from_package(package):
    commands = []

    for modloader, modname, ispkg in pkgutil.iter_modules(package.__path__):
        module = modloader.find_module(modname).load_module(modname)
        commands.extend(load_commands_from_module(module))

        if ispkg:
            commands.extend(load_commands_from_package(module))
    return commands


def load_commands():
    commands = []
    this_module = sys.modules[__name__]

    if this_module.__package__:
        commands = load_commands_from_package(
            sys.modules[this_module.__package__])

    commands.extend(load_commands_from_module(this_module))
    return commands


def main():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(
        title='Commands',
        help="One of the following commands",
        description='use <command> --help to get help on particular command',
        metavar="<command>",
    )
    for command_class in load_commands():
        command = command_class()
        if command.NO_ARGPARSE:
            command_parser = subparsers.add_parser(
                command.NAME,
                help=command.DESCRIPTION,
                add_help=False,
                # Ugly hack to prevent arguments from being parsed as options
                prefix_chars=chr(0),
            )
            command_parser.add_argument('raw_args', nargs=REMAINDER)
        else:
            command_parser = subparsers.add_parser(
                command.NAME, help=command.DESCRIPTION)
        command_parser.set_defaults(command=command)
        command.init_argument_parser(command_parser)

    args = parser.parse_args()
    if not hasattr(args, "command"):
        print(f"Usage: {sys.argv[0]} <command> [<options>]")
    else:
        args.command.args = args
        if 'raw_args' in args:
            args.command.raw_args = args.raw_args
        return args.command.run()


# pylint: disable=protected-access

from uengine.models.storable_model import StorableModel
from .mongo_mock import MongoMockTest

CALLABLE_DEFAULT_VALUE = 4


def callable_default():
    return CALLABLE_DEFAULT_VALUE


class TestModel(StorableModel):

    FIELDS = (
        '_id',
        'field1',
        'field2',
        'field3',
        'callable_default_field'
    )

    VALIDATION_TYPES = {
        "field1": str
    }

    DEFAULTS = {
        'field1': 'default_value',
        'field3': 'required_default_value',
        'callable_default_field': callable_default
    }

    REQUIRED_FIELDS = (
        'field2',
        'field3',
    )

    REJECTED_FIELDS = (
        'field1',
    )

    INDEXES = (
        "field1",
    )


class TestStorableModel(MongoMockTest):

    def setUp(self):
        super().setUp()
        TestModel.destroy_all()

    def tearDown(self):
        TestModel.destroy_all()
        super().tearDown()

    def test_eq(self):
        model = TestModel(field2="mymodel")
        model.save()
        model2 = TestModel.find_one({"field2": "mymodel"})
        self.assertEqual(model, model2)

    def test_reject_on_update(self):
        model = TestModel(field1="original_value",
                          field2="mymodel_reject_test")
        model.save()
        id_ = model._id
        model.update({"field1": "new_value"})
        model = TestModel.find_one({"_id": id_})
        self.assertEqual(model.field1, "original_value")

    def test_update(self):
        model = TestModel(field1="original_value",
                          field2="mymodel_update_test")
        model.save()
        id_ = model._id
        model.update({"field2": "mymodel_updated"})
        model = TestModel.find_one({"_id": id_})
        self.assertEqual(model.field2, "mymodel_updated")

    def test_update_many(self):
        model1 = TestModel(field1="original_value",
                           field2="mymodel_update_test")
        model1.save()
        model2 = TestModel(field1="original_value",
                           field2="mymodel_update_test")
        model2.save()
        model3 = TestModel(field1="do_not_modify",
                           field2="mymodel_update_test")
        model3.save()

        TestModel.update_many({"field1": "original_value"}, {
                              "$set": {"field2": "mymodel_updated"}})
        model1.reload()
        model2.reload()
        model3.reload()

        self.assertEqual(model1.field2, "mymodel_updated")
        self.assertEqual(model2.field2, "mymodel_updated")
        self.assertEqual(model3.field2, "mymodel_update_test")



# pylint: disable=protected-access

from uengine.models.abstract_model import AbstractModel, FieldRequired, InvalidFieldType
from unittest import TestCase

CALLABLE_DEFAULT_VALUE = 4


def callable_default():
    return CALLABLE_DEFAULT_VALUE


class TestModel(AbstractModel):

    FIELDS = (
        '_id',
        'field1',
        'field2',
        'field3',
        'callable_default_field'
    )

    VALIDATION_TYPES = {
        "field1": str
    }

    DEFAULTS = {
        'field1': 'default_value',
        'field3': 'required_default_value',
        'callable_default_field': callable_default
    }

    REQUIRED_FIELDS = (
        'field2',
        'field3',
    )

    REJECTED_FIELDS = (
        'field1',
    )

    INDEXES = (
        "field1",
    )

    AUTO_TRIM_FIELDS = ['field1']


class TestAbstractModel(TestCase):

    def test_init(self):
        model = TestModel(field1='value')
        self.assertEqual(model.field1, 'value')
        model._before_delete()
        model._before_save()

    def test_incomplete(self):
        model = TestModel(field1='value')
        self.assertRaises(FieldRequired, model.save)

    def test_incorrect_index(self):
        with self.assertRaises(TypeError):
            class IncorrectIndexModel(AbstractModel):  # pylint: disable=unused-variable
                INDEXES = (
                    "field1"  # No comma - not a tuple
                )

    def test_invalid_type(self):
        model = TestModel(field1=15, field2="any_value")
        self.assertRaises(InvalidFieldType, model.save)

    def test_merge_on_inheritance(self):
        class Parent(AbstractModel):
            FIELDS = ["pfield"]
            VALIDATION_TYPES = {"pfield": str}
            DEFAULTS = {"pfield": "value"}
            REQUIRED_FIELDS = ["pfield"]
            REJECTED_FIELDS = ["pfield"]
            RESTRICTED_FIELDS = ["pfield"]
            INDEXES = ["pfield"]

        class Child(Parent):
            FIELDS = ["cfield"]
            VALIDATION_TYPES = {"cfield": str}
            DEFAULTS = {"cfield": "value"}
            REQUIRED_FIELDS = ["cfield"]
            REJECTED_FIELDS = ["cfield"]
            RESTRICTED_FIELDS = ["cfield"]
            INDEXES = ["cfield"]

        expected_fields = {"pfield", "cfield"}
        self.assertSetEqual(Child.FIELDS, expected_fields | {"_id"})  # _id comes from AbstractModel
        self.assertSetEqual(Child.REQUIRED_FIELDS, expected_fields)
        self.assertSetEqual(Child.REJECTED_FIELDS, expected_fields)
        self.assertSequenceEqual(sorted(Child.INDEXES), sorted(expected_fields))

        # Unrelated Mixins should also work
        class Mixin:  # pylint: disable=unused-variable
            pass

        class ChildNoOverrides(Mixin, Parent):  # pylint: disable=unused-variable
            pass

    def test_collection_inheritance(self):
        class SemiAbstractModel(AbstractModel):
            pass

        class BaseModel(SemiAbstractModel):
            COLLECTION = "my_collection"

        class Model1(BaseModel):
            pass

        self.assertEqual(SemiAbstractModel.collection, "semi_abstract_model")
        self.assertEqual(BaseModel.collection, "my_collection")
        self.assertEqual(Model1.collection, "model1")

    def test_auto_trim(self):
        t = TestModel(field1="   a   \t", field2="b", field3="c")
        t.save()
        self.assertEqual(t.field1, "a")



from typing import Any
from copy import deepcopy

from bson.objectid import ObjectId
from functools import wraps
from itertools import chain
from pymongo import ASCENDING, DESCENDING, HASHED
from pymongo.errors import OperationFailure
from uengine import ctx
from uengine.errors import ApiError, InvalidFieldType
from .model_hook import ModelHook


def snake_case(name):
    result = ""
    for i, l in enumerate(name):
        if 65 <= ord(l) <= 90:
            if i != 0:
                result += "_"
            result += l.lower()
        else:
            result += l
    return result


def parse_index_key(index_key):
    if index_key.startswith("-"):
        index_key = index_key[1:]
        order = DESCENDING
    elif index_key.startswith("#"):
        index_key = index_key[1:]
        order = HASHED
    else:
        order = ASCENDING
        if index_key.startswith("+"):
            index_key = index_key[1:]
    return index_key, order


class ObjectSaveRequired(Exception):
    pass


class FieldRequired(ApiError):
    def __init__(self, field_name):
        ApiError.__init__(self, "Field \"%s\" is required" %
                          field_name, status_code=400)


def merge_set(attr, new_cls, bases):
    merged = set()
    valid_types = (list, set, frozenset, tuple)
    for cls in chain(bases, [new_cls]):
        cls_attr = getattr(cls, attr, [])
        if not isinstance(cls_attr, valid_types):
            raise TypeError(
                "{} field must be one of: {}".format(attr, valid_types))
        merged.update(cls_attr)

    setattr(new_cls, attr, frozenset(merged))


def merge_tuple(attr, new_cls, bases):
    merged = []
    for cls in chain(bases, [new_cls]):
        cls_attr = getattr(cls, attr, [])
        if not isinstance(cls_attr, (list, set, tuple)):
            raise TypeError(
                "{} field must be of type list, set or tuple".format(attr))
        merged.extend(cls_attr)
    setattr(new_cls, attr, tuple(merged))


def merge_dict(attr, new_cls, bases):
    merged = {}
    for cls in chain(bases, [new_cls]):
        cls_attr = getattr(cls, attr, {})
        if not isinstance(cls_attr, dict):
            raise TypeError("{} field must be a dict".format(attr))
        merged.update(cls_attr)
    setattr(new_cls, attr, merged)


class ModelMeta(type):

    def __new__(mcs, name, bases, dct) -> Any:
        new_cls = super().__new__(mcs, name, bases, dct)

        # First merge mergers config
        merge_dict("_MERGERS", new_cls, bases)

        for attr, merge_func in new_cls._MERGERS.items():
            merge_func(attr, new_cls, bases)

        compatibility_fields = []
        for field in new_cls.FIELDS:
            if field in new_cls.COMPATIBILITY_FIELD_MAP:
                compatibility_fields.append(
                    new_cls.COMPATIBILITY_FIELD_MAP[field])
            else:
                compatibility_fields.append(field)
        new_cls.COMPATIBILITY_FIELDS = frozenset(compatibility_fields)

        new_cls.collection = mcs._get_collection(new_cls, name, bases, dct)

        return new_cls

    @staticmethod
    def _get_collection(model_cls, name, bases, dct):  # pylint: disable=unused-argument
        # Do not inherit collection names from base classes
        if 'COLLECTION' in dct:
            return dct['COLLECTION']
        return snake_case(name)


class AbstractModel(metaclass=ModelMeta):

    FIELDS = [
        "_id",
    ]
    REJECTED_FIELDS = []
    REQUIRED_FIELDS = set()
    RESTRICTED_FIELDS = []
    AUTO_TRIM_FIELDS = []
    KEY_FIELD = None
    DEFAULTS = {}
    VALIDATION_TYPES = {}
    INDEXES = []
    COMPATIBILITY_FIELD_MAP = {}

    AUXILIARY_SLOTS = (
        "AUXILIARY_SLOTS",
        "FIELDS",
        "REJECTED_FIELDS",
        "REQUIRED_FIELDS",
        "RESTRICTED_FIELDS",
        "KEY_FIELD",
        "DEFAULTS",
        "INDEXES",
    )

    _MERGERS = {
        "FIELDS": merge_set,
        "REJECTED_FIELDS": merge_set,
        "REQUIRED_FIELDS": merge_set,
        "RESTRICTED_FIELDS": merge_set,
        "AUXILIARY_SLOTS": merge_set,
        "DEFAULTS": merge_dict,
        "VALIDATION_TYPES": merge_dict,
        "INDEXES": merge_tuple,
    }

    _HOOKS = None

    __hash__ = None
    __slots__ = FIELDS + ["_id", "_hooks"]

    def __init__(self, **kwargs):
        if "_id" not in kwargs:
            self._id = None
        for field, value in kwargs.items():
            if field in self.FIELDS:
                setattr(self, field, value)
        for field in self.FIELDS:
            if field not in kwargs:
                value = self.DEFAULTS.get(field)
                if callable(value):
                    value = value()
                elif hasattr(value, "copy"):
                    value = value.copy()
                elif hasattr(value, "__getitem__"):
                    value = value[:]
                setattr(self, field, value)
        self.__set_initial_state()
        self._hooks = []
        if self._HOOKS:
            for hook_class in self._HOOKS:
                hook_inst = hook_class.on_model_init(self)
                if hook_inst:
                    self._hooks.append(hook_inst)

    def __set_initial_state(self):
        setattr(self, "_initial_state", deepcopy(self.to_dict(self.FIELDS)))

    @classmethod
    def register_model_hook(cls, model_hook_class, *args, **kwargs):
        if not issubclass(model_hook_class, ModelHook):
            raise TypeError("Invalid hook class")
        if cls._HOOKS is None:
            cls._HOOKS = set()
        if model_hook_class not in cls._HOOKS:
            cls._HOOKS.add(model_hook_class)
            model_hook_class.on_hook_register(cls, *args, **kwargs)
            ctx.log.debug("Registered hook %s for model %s",
                          model_hook_class.__name__, cls.__name__)

    @classmethod
    def unregister_model_hook(cls, model_hook_class):
        if cls._HOOKS is None:
            return
        if model_hook_class in cls._HOOKS:
            cls._HOOKS.remove(model_hook_class)
            model_hook_class.on_hook_unregister(cls)

    @classmethod
    def clear_hooks(cls):
        if cls._HOOKS is None:
            return
        for hook_class in cls._HOOKS.copy():
            cls.unregister_model_hook(hook_class)

    def _before_save(self):
        pass

    def _before_validation(self):
        pass

    def _before_delete(self):
        pass

    def _after_save(self, is_new):
        pass

    def _after_delete(self):
        pass

    def _save_to_db(self):
        pass

    def _delete_from_db(self):
        pass

    def invalidate(self):
        pass

    def _validate(self):
        for field in self.missing_fields:
            raise FieldRequired(field)

        for field_name, f_type in self.VALIDATION_TYPES.items():
            if not isinstance(getattr(self, field_name), f_type):
                raise InvalidFieldType(
                    f"field {field_name} must be of type {f_type.__name__}")

    def _reload_from_obj(self, obj):
        for field in self.FIELDS:
            if field == "_id":
                continue
            value = getattr(obj, field)
            setattr(self, field, value)

    def destroy(self, skip_callback=False, invalidate_cache=True):
        if self.is_new:
            return
        if not skip_callback:
            self._before_delete()
        self._delete_from_db()
        if not skip_callback:
            self._after_delete()
        self._id = None
        for hook in self._hooks:
            try:
                hook.on_model_destroy(self)
            except Exception as e:
                ctx.log.error("error executing destroy hook %s on model %s(%s): %s",
                              hook.__class__.__name__, self.__class__.__name__, self._id, e)
        if invalidate_cache:
            self.invalidate()
        return self

    def save(self, skip_callback=False, invalidate_cache=True):
        is_new = self.is_new

        if not skip_callback:
            self._before_validation()
        self._validate()

        # autotrim
        for field in self.AUTO_TRIM_FIELDS:
            value = getattr(self, field)
            try:
                value = value.strip()
                setattr(self, field, value)
            except AttributeError:
                pass

        if not skip_callback:
            self._before_save()
        self._save_to_db()

        for hook in self._hooks:
            try:
                hook.on_model_save(self, is_new)
            except Exception as e:
                ctx.log.error("error executing save hook %s on model %s(%s): %s",
                              hook.__class__.__name__, self.__class__.__name__, self._id, e)

        self.__set_initial_state()
        if invalidate_cache:
            self.invalidate()
        if not skip_callback:
            self._after_save(is_new)

        return self

    def __repr__(self):
        attributes = ["%s=%r" % (a, getattr(self, a))
                      for a in list(self.FIELDS)]
        return '%s(\n    %s\n)' % (self.__class__.__name__, ',\n    '.join(attributes))

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        for field in self.FIELDS:
            if hasattr(self, field):
                if not hasattr(other, field):
                    return False
                if getattr(self, field) != getattr(other, field):
                    return False
            elif hasattr(other, field):
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self, fields=None, include_restricted=False):
        if fields is None:
            fields = list(self.FIELDS)
        result = {}
        for field in fields:
            if field.startswith("_") and field != "_id":
                continue
            if field in self.AUXILIARY_SLOTS:
                continue
            if field in self.RESTRICTED_FIELDS and not include_restricted:
                continue
            try:
                value = getattr(self, field)
            except AttributeError:
                continue
            if callable(value):
                continue
            result[field] = value
        return result

    @classmethod
    def from_data(cls, **data):
        return cls(**data)

    @property
    def is_complete(self):
        return len(self.missing_fields) == 0

    @property
    def is_new(self):
        return not (hasattr(self, "_id") and isinstance(self._id, ObjectId))

    @property
    def missing_fields(self):
        mfields = []
        for field in self.REQUIRED_FIELDS:
            if not hasattr(self, field) or getattr(self, field) in ["", None]:
                mfields.append(field)
        return mfields

    @classmethod
    def _get_possible_databases(cls):
        return [ctx.db.meta]

    @classmethod
    def ensure_indexes(cls, loud=False, overwrite=False):  # pylint: disable=too-many-branches
        if not isinstance(cls.INDEXES, (list, tuple)):
            raise TypeError("INDEXES field must be of type list or tuple")

        for index in cls.INDEXES:
            if isinstance(index, str):
                index = [index]
            keys = []
            options = {"sparse": False}

            for sub_index in index:
                if isinstance(sub_index, str):
                    keys.append(parse_index_key(sub_index))
                else:
                    for key, value in sub_index.items():
                        options[key] = value
            if loud:
                ctx.log.debug(
                    "Creating index with options: %s, %s", keys, options)

            for db in cls._get_possible_databases():
                try:
                    db.conn[cls.collection].create_index(keys, **options)
                except OperationFailure as e:
                    if e.details.get("codeName") == "IndexOptionsConflict" or e.details.get("code") == 85:
                        if overwrite:
                            if loud:
                                ctx.log.debug(
                                    "Dropping index %s as conflicting", keys)
                            db.conn[cls.collection].drop_index(keys)
                            if loud:
                                ctx.log.debug(
                                    "Creating index with options: %s, %s", keys, options)
                            db.conn[cls.collection].create_index(
                                keys, **options)
                        else:
                            ctx.log.error(
                                "Index %s conflicts with an existing one, use overwrite param to fix it", keys
                            )


def save_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        this = args[0]
        if this.is_new:
            raise ObjectSaveRequired("This object must be saved first")
        return func(*args, **kwargs)
    return wrapper
